@@ -160,6 +160,20 @@ class TestAtoms(BaseTest):
 
         assert cp.power(-1, 2).value == 1
 
+    # Test the xexp class
+    def test_xexp(self) -> None:
+        # Test for positive x
+        x = Variable(pos=True)
+        atom = cp.xexp(x)
+        self.assertEqual(atom.curvature, s.CONVEX)
+        self.assertEqual(atom.sign, s.NONNEG)
+
+        # Test for negative x
+        x = Variable(neg=True)
+        atom = cp.xexp(x)
+        self.assertNotEqual(atom.curvature, s.CONCAVE)
+        self.assertEqual(atom.sign, s.NONPOS)
+
     # Test the geo_mean class.
     def test_geo_mean(self) -> None:
         atom = cp.geo_mean(self.x)
@@ -336,8 +350,10 @@ class TestAtoms(BaseTest):
         # Invalid axis.
         with self.assertRaises(Exception) as cm:
             cp.max(self.x, axis=4)
-        self.assertEqual(str(cm.exception),
-                         "Invalid argument for axis.")
+        self.assertEqual(str(cm.exception), "Invalid argument for axis.")
+        with self.assertRaises(ValueError) as cm:
+            cp.max(self.x, self.x)  # a common erroneous use-case
+        self.assertEqual(str(cm.exception), cp.max.__EXPR_AXIS_ERROR__)
 
     def test_min(self) -> None:
         """Test min.
@@ -357,8 +373,10 @@ class TestAtoms(BaseTest):
         # Invalid axis.
         with self.assertRaises(Exception) as cm:
             cp.min(self.x, axis=4)
-        self.assertEqual(str(cm.exception),
-                         "Invalid argument for axis.")
+        self.assertEqual(str(cm.exception), "Invalid argument for axis.")
+        with self.assertRaises(ValueError) as cm:
+            cp.min(self.x, self.x)  # a common erroneous use-case
+        self.assertEqual(str(cm.exception), cp.min.__EXPR_AXIS_ERROR__)
 
     # Test sign logic for maximum.
     def test_maximum_sign(self) -> None:
@@ -818,7 +836,7 @@ class TestAtoms(BaseTest):
         self.assertEqual(str(cm.exception),
                          "The arguments to conv must resolve to vectors.")
 
-    def test_kron(self) -> None:
+    def test_kron_expr(self) -> None:
         """Test the kron atom.
         """
         a = np.ones((3, 2))
@@ -830,9 +848,9 @@ class TestAtoms(BaseTest):
         expr = cp.kron(a, b)
         assert expr.is_nonpos()
         with self.assertRaises(Exception) as cm:
-            cp.kron(self.x, -1)
+            cp.kron(self.x, self.x)
         self.assertEqual(str(cm.exception),
-                         "The first argument to kron must be constant.")
+                         "At least one argument to kron must be constant.")
 
     def test_partial_optimize_dcp(self) -> None:
         """Test DCP properties of partial optimize.
@@ -1168,17 +1186,122 @@ class TestAtoms(BaseTest):
         result = prob.solve(solver=cp.SCS)
         assert np.isclose(result, true_val.sum(), atol=1e0)
 
-    def test_log_sum_exp(self) -> None:
-        """Test log_sum_exp sign.
+    def test_partial_trace(self) -> None:
         """
-        # Test for non-negative x
-        x = Variable(nonneg=True)
-        atom = cp.log_sum_exp(x)
-        self.assertEqual(atom.curvature, s.CONVEX)
-        self.assertEqual(atom.sign, s.NONNEG)
+        Test partial_trace atom.
+        rho_ABC = rho_A \\otimes rho_B \\otimes rho_C.
+        Here \\otimes signifies Kronecker product.
+        Each rho_i is normalized, i.e. Tr(rho_i) = 1.
+        """
+        # Set random state.
+        np.random.seed(1)
 
-        # Test for non-positive x
-        x = Variable(nonpos=True)
-        atom = cp.log_sum_exp(x)
-        self.assertEqual(atom.curvature, s.CONVEX)
-        self.assertEqual(atom.sign, s.UNKNOWN)
+        # Generate five test cases
+        rho_A = np.random.random((4, 4)) + 1j*np.random.random((4, 4))
+        rho_A /= np.trace(rho_A)
+        rho_B = np.random.random((3, 3)) + 1j*np.random.random((3, 3))
+        rho_B /= np.trace(rho_B)
+        rho_C = np.random.random((2, 2)) + 1j*np.random.random((2, 2))
+        rho_C /= np.trace(rho_C)
+        rho_AB = np.kron(rho_A, rho_B)
+        rho_AC = np.kron(rho_A, rho_C)
+
+        # Construct a cvxpy Variable with value equal to rho_A \otimes rho_B \otimes rho_C.
+        temp = np.kron(rho_AB, rho_C)
+        rho_ABC = cp.Variable(shape=temp.shape, complex=True)
+        rho_ABC.value = temp
+
+        # Try to recover simpler tensors products by taking partial traces of
+        # more complicated tensors.
+        rho_AB_test = cp.partial_trace(rho_ABC, [4, 3, 2], axis=2)
+        rho_AC_test = cp.partial_trace(rho_ABC, [4, 3, 2], axis=1)
+        rho_A_test = cp.partial_trace(rho_AB_test, [4, 3], axis=1)
+        rho_B_test = cp.partial_trace(rho_AB_test, [4, 3], axis=0)
+        rho_C_test = cp.partial_trace(rho_AC_test, [4, 2], axis=0)
+
+        # See if the outputs of partial_trace are correct
+        assert np.allclose(rho_AB_test.value, rho_AB)
+        assert np.allclose(rho_AC_test.value, rho_AC)
+        assert np.allclose(rho_A_test.value, rho_A)
+        assert np.allclose(rho_B_test.value, rho_B)
+        assert np.allclose(rho_C_test.value, rho_C)
+
+    def test_partial_trace_exceptions(self) -> None:
+        """Test exceptions raised by partial trace.
+        """
+        X = cp.Variable((4, 3))
+        with self.assertRaises(ValueError) as cm:
+            cp.partial_trace(X, dims=[2, 3], axis=0)
+        self.assertEqual(str(cm.exception),
+                         "Only supports square matrices.")
+
+        X = cp.Variable((6, 6))
+        with self.assertRaises(ValueError) as cm:
+            cp.partial_trace(X, dims=[2, 3], axis=-1)
+        self.assertEqual(str(cm.exception),
+                         "Invalid axis argument, should be between 0 and 2, got -1.")
+
+        X = cp.Variable((6, 6))
+        with self.assertRaises(ValueError) as cm:
+            cp.partial_trace(X, dims=[2, 4], axis=0)
+        self.assertEqual(str(cm.exception),
+                         "Dimension of system doesn't correspond to dimension of subsystems.")
+
+    def test_partial_transpose(self) -> None:
+        """
+        Test out the partial_transpose atom.
+        rho_ABC = rho_A \\otimes rho_B \\otimes rho_C.
+        Here \\otimes signifies Kronecker product.
+        Each rho_i is normalized, i.e. Tr(rho_i) = 1.
+        """
+        # Set random state.
+        np.random.seed(1)
+
+        # Generate three test cases
+        rho_A = np.random.random((8, 8)) + 1j*np.random.random((8, 8))
+        rho_A /= np.trace(rho_A)
+        rho_B = np.random.random((6, 6)) + 1j*np.random.random((6, 6))
+        rho_B /= np.trace(rho_B)
+        rho_C = np.random.random((4, 4)) + 1j*np.random.random((4, 4))
+        rho_C /= np.trace(rho_C)
+
+        rho_TC = np.kron(np.kron(rho_A, rho_B), rho_C.T)
+        rho_TB = np.kron(np.kron(rho_A, rho_B.T), rho_C)
+        rho_TA = np.kron(np.kron(rho_A.T, rho_B), rho_C)
+
+        # Construct a cvxpy Variable with value equal to rho_A \otimes rho_B \otimes rho_C.
+        temp = np.kron(np.kron(rho_A, rho_B), rho_C)
+        rho_ABC = cp.Variable(shape=temp.shape, complex=True)
+        rho_ABC.value = temp
+
+        # Try to recover simpler tensors products by taking partial transposes of
+        # more complicated tensors.
+        rho_TC_test = cp.partial_transpose(rho_ABC, [8, 6, 4], axis=2)
+        rho_TB_test = cp.partial_transpose(rho_ABC, [8, 6, 4], axis=1)
+        rho_TA_test = cp.partial_transpose(rho_ABC, [8, 6, 4], axis=0)
+
+        # See if the outputs of partial_transpose are correct
+        assert np.allclose(rho_TC_test.value, rho_TC)
+        assert np.allclose(rho_TB_test.value, rho_TB)
+        assert np.allclose(rho_TA_test.value, rho_TA)
+
+    def test_partial_transpose_exceptions(self) -> None:
+        """Test exceptions raised by partial trace.
+        """
+        X = cp.Variable((4, 3))
+        with self.assertRaises(ValueError) as cm:
+            cp.partial_transpose(X, dims=[2, 3], axis=0)
+        self.assertEqual(str(cm.exception),
+                         "Only supports square matrices.")
+
+        X = cp.Variable((6, 6))
+        with self.assertRaises(ValueError) as cm:
+            cp.partial_transpose(X, dims=[2, 3], axis=-1)
+        self.assertEqual(str(cm.exception),
+                         "Invalid axis argument, should be between 0 and 2, got -1.")
+
+        X = cp.Variable((6, 6))
+        with self.assertRaises(ValueError) as cm:
+            cp.partial_transpose(X, dims=[2, 4], axis=0)
+        self.assertEqual(str(cm.exception),
+                         "Dimension of system doesn't correspond to dimension of subsystems.")
